import JSON5 from 'json5';
import { ConfiguredVueProject, InferredVueProject, createFullAnalyzer, getComponentName } from '@vuedx/analyze';
import { VueTextDocument, isVirtualFile, getContainingFile, DocumentStore, asUri, isVueFile, MODULE_SELECTOR, parseVirtualFileName, VIRTUAL_FILENAME_SEPARATOR, INTERNAL_MODULE_SELECTOR, RENDER_SELECTOR, isNumber } from '@vuedx/vue-virtual-textdocument';
import Path from 'path';
import { URI } from 'vscode-uri';
import QuickLRU from 'quick-lru';
import { isRootNode, isElementNode, isTextNode, isInterpolationNode, isCommentNode, traverseEvery, isSimpleExpressionNode, isDirectiveNode, isAttributeNode, isComponentNode, traverseFast, isPlainElementNode } from '@vuedx/template-ast-types';

function wrapInTrace(context, target) {
    return target;
}
function wrapFn(context, target) {
    return target;
}

const PATCHED_METHODS = Symbol('Vue Patched Methods');
function tryPatchMethod(target, methodName, createOverride) {
    const patched = target[PATCHED_METHODS] || [];
    if (patched.includes(methodName))
        return;
    let fn = target[methodName];
    if (typeof fn === 'function') {
        try {
            fn = fn.bind(target);
        }
        catch (_a) {
            // - ignore
        }
    }
    target[methodName] = createOverride(fn);
    patched.push(methodName);
    target[PATCHED_METHODS] = patched;
}

function getLastNumberFromVersion(version) {
    const parts = version.split(/[^0-9]+/);
    const ver = parts.pop();
    return Number(ver);
}
class ProxyDocumentStore extends DocumentStore {
    get(fileNameOrUri) {
        return super.get(asUri(fileNameOrUri));
    }
    has(fileNameOrUri) {
        return super.has(asUri(fileNameOrUri));
    }
    set(fileNameOrUri, document) {
        return super.set(asUri(fileNameOrUri), document);
    }
    delete(fileNameOrUri) {
        return super.delete(asUri(fileNameOrUri));
    }
}
function getConfig(config = {}) {
    return {
        ...config,
        features: {
            diagnostics: ['semantic', 'suggestion', 'syntactic'],
            organizeImports: true,
            quickInfo: true,
            rename: true,
            refactor: true,
            goto: true,
            ...config.features,
        },
        directories: [
            { kind: 'component', name: 'components', path: 'src/components' },
        ],
    };
}
class PluginContext {
    constructor(typescript) {
        this.typescript = typescript;
        this._config = getConfig();
        this._externalFiles = new WeakMap();
        this._vueProjects = [];
        this.lastCleanupAt = Date.now();
        this.store = new ProxyDocumentStore((uri) => {
            var _a;
            const fileName = URI.parse(uri).fsPath;
            const content = (_a = this.typescript.sys.readFile(fileName)) !== null && _a !== void 0 ? _a : '';
            return this.createVueDocument(fileName, content);
        }, (uri) => {
            const fileName = URI.parse(uri).fsPath;
            return URI.file(this.projectService.toPath(fileName)).toString();
        });
    }
    get config() {
        return this._config;
    }
    get serviceHost() {
        return this._serverHost;
    }
    get projectService() {
        return this._projectService;
    }
    log(message) {
        if (this.projectService != null) {
            this.projectService.logger.info(`Vue.js:: ${message}`);
        }
    }
    getVueVersion(fileName) {
        return '3.0.0';
    }
    getExternalFiles(project) {
        var _a;
        return (_a = this._externalFiles.get(project)) !== null && _a !== void 0 ? _a : [];
    }
    forEachTSProject(fn) {
        this.projectService.configuredProjects.forEach(fn);
        this.projectService.inferredProjects.forEach(fn);
        this.projectService.externalProjects.forEach(fn);
    }
    getVueProjectForFile(fileName, ensure) {
        var _a, _b, _c;
        let project = (_b = (_a = this._vueProjects.find(({ project }) => fileName.startsWith(project.rootDir + '/'))) === null || _a === void 0 ? void 0 : _a.project) !== null && _b !== void 0 ? _b : null;
        if (project === null && ensure === true) {
            const packageFile = this.typescript.findConfigFile(fileName, this.typescript.sys.fileExists, 'package.json');
            const configFile = this.typescript.findConfigFile(fileName, this.typescript.sys.fileExists, 'vueconfig.json');
            const rootDir = Path.posix.dirname((_c = packageFile !== null && packageFile !== void 0 ? packageFile : configFile) !== null && _c !== void 0 ? _c : fileName);
            const fileNames = this.serviceHost.readDirectory(rootDir, ['.tsx', '.jsx', '.ts', '.js'], ['node_modules']);
            const tryRequire = (fileName) => {
                var _a;
                try {
                    if (Path.posix.basename(fileName) === 'vueconfig.json') {
                        const contents = (_a = this.serviceHost.readFile(fileName)) !== null && _a !== void 0 ? _a : '';
                        return JSON5.parse(contents);
                    }
                    // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
                    delete require.cache[fileName];
                    return require(fileName);
                }
                catch (error) {
                    return {};
                }
            };
            const newProject = configFile != null
                ? new ConfiguredVueProject(rootDir, packageFile, packageFile != null ? tryRequire(packageFile) : {}, configFile, tryRequire(configFile))
                : new InferredVueProject(rootDir, packageFile, packageFile != null ? tryRequire(packageFile) : {});
            newProject.setFileNames(fileNames);
            const disposables = [
                () => {
                    this._vueProjects.splice(this._vueProjects.findIndex((item) => item.project === newProject), 1);
                },
            ];
            const dispose = () => disposables.forEach((fn) => fn());
            const projectDirWatcher = this.serviceHost.watchDirectory(rootDir, (fileName) => {
                if (fileName.endsWith('/vueconfig.json')) {
                    if (newProject.kind === 'inferred') {
                        dispose();
                        reload();
                        return;
                    }
                }
                if (!/\.(vue|ts|tsx|js|jsx)$/.test(fileName))
                    return;
                if (this.serviceHost.fileExists(fileName)) {
                    if (!newProject.fileNames.includes(fileName)) {
                        newProject.setFileNames([...newProject.fileNames, fileName]);
                    }
                }
                else {
                    if (newProject.fileNames.includes(fileName)) {
                        newProject.setFileNames(newProject.fileNames.filter((f) => f !== fileName));
                    }
                }
            });
            disposables.push(() => projectDirWatcher.close());
            const reload = () => {
                this.forEachTSProject((tsProject) => {
                    const dir = tsProject.getCurrentDirectory();
                    if (dir.startsWith(newProject.rootDir) ||
                        newProject.rootDir.startsWith(dir)) {
                        this.getExternalFiles(tsProject).forEach((fileName) => {
                            const document = this.store.get(fileName);
                            if (document != null) {
                                document.markDirty();
                                document.all().forEach((doc) => {
                                    if ([
                                        'javascript',
                                        'typescript',
                                        'javascriptreact',
                                        'typescriptreact',
                                    ].includes(doc.languageId)) {
                                        triggerFileUpdate(this, doc.fsPath);
                                    }
                                });
                            }
                        });
                        tsProject.markAsDirty();
                        tsProject.refreshDiagnostics();
                    }
                });
            };
            if (configFile != null) {
                const configFileWatcher = this.serviceHost.watchFile(configFile, (_, event) => {
                    if (event === this.typescript.FileWatcherEventKind.Deleted) {
                        dispose();
                    }
                    else {
                        newProject.config = tryRequire(configFile);
                        newProject.markDirty();
                    }
                    reload();
                });
                disposables.push(() => {
                    configFileWatcher.close();
                });
            }
            if (packageFile != null) {
                const packageFileWatcher = this.serviceHost.watchFile(packageFile, (_, event) => {
                    if (event === this.typescript.FileWatcherEventKind.Deleted) {
                        dispose();
                    }
                    else {
                        newProject.packageJSON = {
                            dependencies: {},
                            devDependencies: {},
                            ...tryRequire(packageFile),
                        };
                        newProject.markDirty();
                    }
                });
                disposables.push(() => packageFileWatcher.close());
            }
            this._vueProjects.push({ project: newProject, dispose });
            project = newProject;
        }
        return project;
    }
    disposeUnusedProjects(force = false) {
        if (!force &&
            (Date.now() - this.lastCleanupAt > 5 * 60 * 1000 ||
                this._vueProjects.length === 0)) {
            return;
        }
        this.lastCleanupAt = Date.now();
        const usedProjects = new Set();
        // TODO: detect unused projects
        this._vueProjects.forEach(({ project, dispose }) => {
            if (!usedProjects.has(project)) ;
        });
    }
    createVueDocument(fileName, content) {
        const uri = URI.file(fileName).toString();
        const document = VueTextDocument.create(uri, 'vue', 0, content, {
            getGlobalComponents: wrapFn('getGlobalComponents', () => {
                const project = this.getVueProjectForFile(fileName, true);
                return project.kind === 'inferred'
                    ? project.components
                    : project.globalComponents;
            }),
        });
        this.store.set(uri, document);
        return document;
    }
    error(error) {
        var _a;
        if (this.projectService != null) {
            this.projectService.logger.msg(`Vue.js:: ${error.message} ${(_a = error.stack) !== null && _a !== void 0 ? _a : ''}`, this.typescript.server.Msg.Err);
        }
    }
    tryCreateScriptInfo(fileName) {
        try {
            if (isVirtualFile(fileName)) {
                const vueFileName = getContainingFile(fileName);
                this.store.get(vueFileName);
                const scriptInfo = this.projectService.getOrCreateScriptInfoForNormalizedPath(this.typescript.server.toNormalizedPath(vueFileName), false);
                if (scriptInfo != null) {
                    patchScriptInfo(this, scriptInfo);
                }
            }
        }
        catch (_a) { }
    }
    load(info) {
        this.log(`Loading Vue plugin: ${info.project.getProjectName()}`);
        this._serverHost = info.serverHost;
        this._projectService = info.project.projectService;
        this.setConfig(info.config);
        patchProjectService(this);
        patchServiceHost(this);
        patchLanguageServiceHost(this, info.languageServiceHost);
    }
    setConfig(config) {
        this._config = getConfig(config);
    }
}
function patchProjectService(context) {
    patchExtraFileExtensions(context);
}
function patchExtraFileExtensions(context) {
    var _a;
    const extraFileExtensions = [
        {
            extension: 'vue',
            isMixedContent: false,
            scriptKind: context.typescript.ScriptKind.Deferred,
        },
    ];
    tryPatchMethod(context.projectService, 'setHostConfiguration', (setHostConfiguration) => {
        context.log(`[patch] Add support for vue extension. (ProjectService)`);
        return wrapFn('setHostConfiguration', (args) => {
            const current = context.projectService
                .hostConfiguration
                .extraFileExtensions;
            if (args.extraFileExtensions != null) {
                args.extraFileExtensions.push(...extraFileExtensions);
                context.log(`extraFileExtensions: ${JSON.stringify(args.extraFileExtensions)}`);
            }
            else if (current == null ||
                !current.some((ext) => ext.extension === 'vue')) {
                args.extraFileExtensions = [...extraFileExtensions];
            }
            return setHostConfiguration(args);
        });
    });
    if (((_a = context.projectService
        .hostConfiguration.extraFileExtensions) === null || _a === void 0 ? void 0 : _a.some((ext) => ext.extension === 'vue')) === true) {
        return;
    }
    // Enable .vue after enhancing the language server.
    context.projectService.setHostConfiguration({ extraFileExtensions: [] });
}
function patchLanguageServiceHost(context, languageServiceHost) {
    patchGetScriptFileNames(context, languageServiceHost);
    patchModuleResolution(context, languageServiceHost);
    tryPatchMethod(languageServiceHost, 'getCompilationSettings', (getCompilationSettings) => () => {
        const settings = getCompilationSettings();
        settings.jsx = context.typescript.JsxEmit.Preserve;
        if (settings.allowJs !== false) {
            settings.allowJs = true;
        }
        if (settings.checkJs !== false) {
            settings.checkJs = true;
        }
        return settings;
    });
}
function patchGetScriptFileNames(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'getScriptFileNames', (getScriptFileNames) => {
        context.log(`[patch] Override getScriptFileNames to expand .vue files to virtual files. (LanguageServerHost)`);
        return wrapFn('getScriptFileNames', () => {
            const fileNames = new Set();
            const vueFiles = new Set();
            getScriptFileNames()
                .map((fileName) => isVirtualFile(fileName) ? getContainingFile(fileName) : fileName)
                .forEach((fileName) => {
                if (isVueFile(fileName)) {
                    const document = context.store.get(fileName);
                    if (document != null) {
                        vueFiles.add(fileName);
                        fileNames.add(document.getDocumentFileName(MODULE_SELECTOR));
                    }
                }
                else {
                    fileNames.add(fileName);
                }
            });
            if (vueFiles.size > 0) {
                const files = Array.from(vueFiles);
                const projects = new Set(files
                    .map((fileName) => context.projectService.getDefaultProjectForFile(context.typescript.server.toNormalizedPath(fileName), false))
                    .filter(Boolean));
                projects.forEach((project) => {
                    context._externalFiles.set(project, files);
                    if (project instanceof context.typescript.server.ConfiguredProject) {
                        const vueProject = context.getVueProjectForFile(project.getCurrentDirectory() + '/placeholder.vue', true);
                        vueProject.setFileNames(files);
                    }
                });
            }
            return Array.from(fileNames);
        });
    });
}
function patchFileExists(context) {
    tryPatchMethod(context.serviceHost, 'fileExists', (fileExists = context.typescript.sys.fileExists) => {
        context.log(`[patch] Override fileExists to check containing file for virtual files. (ServiceHost)`);
        return wrapFn('fileExists', (fileName) => {
            if (isVirtualFile(fileName)) {
                const document = context.store.get(getContainingFile(fileName));
                const result = parseVirtualFileName(fileName);
                return (document != null &&
                    result != null &&
                    fileName === document.getDocumentFileName(result.selector));
            }
            return fileExists(fileName);
        });
    });
}
function patchReadFile(context) {
    tryPatchMethod(context.serviceHost, 'readFile', (readFile) => {
        context.log(`[patch] Override readFile to check containing file for virtual files. (ServiceHost)`);
        return wrapFn('readFile', (fileName, encoding) => {
            var _a;
            if (isVirtualFile(fileName)) {
                context.log(`host.readFile("${fileName}")`);
                const document = (_a = context.store
                    .get(getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName);
                if (document != null)
                    return document.getText();
                else
                    return;
            }
            return readFile != null
                ? readFile(fileName, encoding)
                : context.typescript.sys.readFile(fileName, encoding);
        });
    });
}
function patchReadDirectory(context) {
    tryPatchMethod(context.serviceHost, 'readDirectory', (readDirectory) => {
        return wrapFn('readDirectory', (path, extensions, exclude, include, depth) => {
            if (extensions != null) {
                extensions = [...extensions, '.vue'];
            }
            return readDirectory(path, extensions, exclude, include, depth);
        });
    });
}
function patchModuleResolution(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'resolveModuleNames', (resolveModuleNames) => {
        context.log(`[patch] Override resolveModuleNames to resolve imports from .vue files. (LanguageServerHost)`);
        return wrapFn('resolveModuleNames', (moduleNames, containingFile, reusedNames, redirectedReferences, options) => {
            if (isVueFile(containingFile))
                throw new Error('A .vue file should not be part of TS program.');
            const newModuleNames = moduleNames.map((moduleName) => isVueFile(moduleName)
                ? moduleName + VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR
                : moduleName.endsWith('.vue?internal')
                    ? moduleName.replace(/\?internal$/, '') +
                        VIRTUAL_FILENAME_SEPARATOR +
                        INTERNAL_MODULE_SELECTOR
                    : moduleName);
            // TODO: Support paths mapped to .vue files, if needed.
            const result = resolveModuleNames != null
                ? resolveModuleNames(newModuleNames, containingFile, reusedNames, redirectedReferences, options)
                : [];
            result.forEach((resolved) => {
                if (resolved != null && isVirtualFile(resolved.resolvedFileName)) {
                    context.tryCreateScriptInfo(resolved.resolvedFileName); // Trigger .vue script info creation.
                }
            });
            return result;
        });
    });
}
function patchServiceHost(context) {
    patchWatchFile(context);
    patchWatchDirectory(context);
    patchFileExists(context);
    patchReadFile(context);
    patchReadDirectory(context);
}
function patchWatchFile(context) {
    tryPatchMethod(context.serviceHost, 'watchFile', (watchFile) => {
        context.log(`[patch] Override watchFile to watch virtual files. (ServiceHost)`);
        return wrapFn('watchFile', (fileName, callback, pollingInterval, options) => {
            if (isVirtualFile(fileName)) {
                return watchFile(getContainingFile(fileName), (_, event) => {
                    callback(fileName, event);
                }, pollingInterval, options);
            }
            return watchFile(fileName, callback, pollingInterval, options);
        });
    });
}
function patchWatchDirectory(context) {
    tryPatchMethod(context.serviceHost, 'watchDirectory', (watchDirectory) => {
        context.log(`[patch] Override watchFile to watch virtual files. (ServiceHost)`);
        return wrapFn('watchDirectory', (fileName, callback, recursive, options) => {
            return watchDirectory(fileName, (fileName) => {
                if (isVueFile(fileName)) {
                    // TODO: Ensure this does mess with fileName case-sensitivity
                    const document = context.store.get(fileName);
                    if (document != null) {
                        callback(document.getDocumentFileName('_module'));
                    }
                }
                else {
                    callback(fileName);
                }
            }, recursive, options);
        });
    });
}
function triggerFileUpdate(context, fileName) {
    const scriptInfo = context.projectService.getScriptInfo(fileName);
    if (scriptInfo != null) {
        // @ts-expect-error - internal method but it's better for performance compared to it's public counter part `reloadFromFile()`.
        scriptInfo.delayReloadNonMixedContentFile();
    }
}
function patchScriptInfo(context, scriptInfo) {
    if (scriptInfo == null)
        throw new Error('ScriptInfo is required.');
    tryPatchMethod(scriptInfo, 'editContent', (editContent) => {
        context.log(`[patch] Override editContent() of "${scriptInfo.fileName}" to sync virtual files. (ScriptInfo)`);
        return wrapFn('editContent', (start, end, newText) => {
            const document = context.store.get(scriptInfo.fileName);
            if (document == null)
                throw new Error('VueTextDocument should exist for every ScriptInfo.');
            const range = {
                start: document.positionAt(start),
                end: document.positionAt(end),
            };
            editContent(start, end, newText);
            VueTextDocument.update(document, [{ range, text: newText }], getLastNumberFromVersion(scriptInfo.getLatestVersion()))
                .all()
                .forEach((document) => triggerFileUpdate(context, document.fsPath));
        });
    });
}

function findTemplateNodeAt(ast, position) {
    return findTemplateNodeFor(ast, position, position);
}
function findTemplateElementNodeAt(ast, position) {
    var _a, _b;
    const result = findTemplateNodeFor(ast, position, position);
    while (result.ancestors.length > 0) {
        if (isRootNode(result.node) ||
            isElementNode(result.node) ||
            isTextNode(result.node) ||
            isInterpolationNode(result.node) ||
            isCommentNode(result.node)) {
            break;
        }
        result.node = (_b = (_a = result.ancestors.pop()) === null || _a === void 0 ? void 0 : _a.node) !== null && _b !== void 0 ? _b : null;
    }
    return result;
}
function findTemplateNodeFor(ast, start, end) {
    const found = {
        node: null,
        ancestors: [],
    };
    traverseEvery(ast, (node, ancestors) => {
        if (node.loc.start.offset <= start && end <= node.loc.end.offset) {
            found.node = node;
            found.ancestors = ancestors.slice();
            return true;
        }
        return false;
    });
    return found;
}
function findTemplateNodesIn(ast, start, end) {
    var _a, _b;
    if (start === end) {
        const a = findTemplateElementNodeAt(ast, start);
        return a.node != null ? [a.node] : [];
    }
    const a = findTemplateElementNodeAt(ast, start);
    const b = findTemplateElementNodeAt(ast, end);
    if (a.node == null || b.node == null)
        return [];
    if (a.node === b.node)
        return [a.node];
    const pa = (_a = a.ancestors.pop()) === null || _a === void 0 ? void 0 : _a.node;
    const pb = (_b = b.ancestors.pop()) === null || _b === void 0 ? void 0 : _b.node;
    if (pa == null || pb == null)
        return [];
    if (pa === b.node)
        return [pa];
    if (pb === a.node)
        return [pb];
    if (pa === pb && isElementNode(pa)) {
        return pa.children.slice(pa.children.indexOf(a.node), 1 + pa.children.indexOf(b.node));
    }
    return [];
}

function createCachedAnalyzer() {
    const cache = new QuickLRU({ maxSize: 1000 });
    const analyzer = createFullAnalyzer([], {
        babel: { plugins: ['typescript', 'jsx'] },
    });
    return (document) => {
        var _a;
        const key = `${document.version}::${document.fsPath}`;
        const info = (_a = cache.get(key)) !== null && _a !== void 0 ? _a : analyzer.analyze(document.getText(), document.fsPath);
        cache.set(key, info);
        return info;
    };
}
// eslint-disable-next-line @typescript-eslint/explicit-function-return-type
function createServerHelper(context, languageService) {
    const getComponentInfo = createCachedAnalyzer();
    function findNodeAtPosition(fileName, position) {
        const document = getRenderDoc(fileName);
        if ((document === null || document === void 0 ? void 0 : document.ast) != null)
            return {
                document: document,
                ...findTemplateNodeAt(document.ast, position),
            };
        else
            return { node: null, ancestors: [], document: null };
    }
    function getDocument(fileName) {
        var _a, _b;
        return isVirtualFile(fileName)
            ? (_b = (_a = context.store.get(getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName)) !== null && _b !== void 0 ? _b : null : context.store.get(fileName);
    }
    function getVueDocument(fileName) {
        return isVueFile(fileName)
            ? context.store.get(fileName)
            : isVirtualFile(fileName)
                ? context.store.get(getContainingFile(fileName))
                : null;
    }
    function getDocumentAt(fileName, position) {
        const document = context.store.get(fileName);
        const block = document === null || document === void 0 ? void 0 : document.blockAt(position);
        if (block == null || document == null)
            return null;
        if (block.type === 'template') {
            return document.getDocument(RENDER_SELECTOR);
        }
        const selector = document.getBlockSelector(block);
        if (selector == null)
            return null;
        return document.getDocument(selector);
    }
    function isRenderFunctionDocument(document) {
        var _a;
        return ((_a = document) === null || _a === void 0 ? void 0 : _a.selector.type) === RENDER_SELECTOR;
    }
    function isRenderFunctionFileName(fileName) {
        var _a;
        return (isVirtualFile(fileName) &&
            ((_a = parseVirtualFileName(fileName)) === null || _a === void 0 ? void 0 : _a.selector.type) === RENDER_SELECTOR);
    }
    function isVueModuleFileName(fileName) {
        var _a;
        return (isVirtualFile(fileName) &&
            ((_a = parseVirtualFileName(fileName)) === null || _a === void 0 ? void 0 : _a.selector.type) === MODULE_SELECTOR);
    }
    function getRenderDoc(fileName) {
        var _a;
        return (_a = getVueDocument(fileName)) === null || _a === void 0 ? void 0 : _a.getDocument('_render');
    }
    function getTextSpan(document, span) {
        if (isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result != null)
                return { start: result.offset, length: result.length };
        }
        return span;
    }
    function getResolvedModule(fileName, importSource) {
        const program = languageService.getProgram();
        const sourceFile = program === null || program === void 0 ? void 0 : program.getSourceFile(fileName);
        const resolvedModules = sourceFile === null || sourceFile === void 0 ? void 0 : sourceFile.resolvedModules;
        return resolvedModules === null || resolvedModules === void 0 ? void 0 : resolvedModules.get(importSource);
    }
    return {
        findNodeAtPosition,
        findTemplateElementNodeAt,
        findTemplateNodeAt,
        findTemplateNodeFor,
        findTemplateNodesIn,
        getComponentInfo,
        getDocument,
        getDocumentAt,
        getRenderDoc,
        getTextSpan,
        getVueDocument,
        isRenderFunctionDocument,
        isRenderFunctionFileName,
        isVueModuleFileName,
        getResolvedModule,
    };
}
function isNotNull(value) {
    return value != null;
}
function computeIdentifierReplacement(source, identifer) {
    const RE = new RegExp(`\\b${identifer}\\b`);
    const match = RE.exec(source);
    if (match == null)
        return { prefixText: source, suffixText: '' };
    return {
        prefixText: source.substr(0, match.index),
        suffixText: source.substr(match.index + match[0].length),
    };
}
function getPaddingLength(source, offset = 0) {
    source = source.substr(offset);
    const match = /^[\s\r\n]+/m.exec(source);
    return match != null ? match[0].length : 0;
}
function getFilenameForNewComponent(context, directory, usedNames = new Set()) {
    let name = 'Component';
    let fileName = Path.posix.join(directory, 'Component.vue');
    let index = 0;
    if (context.serviceHost.directoryExists(directory)) {
        const files = new Set(context.serviceHost.readDirectory(directory, ['.vue'], [], [], 1));
        while (files.has(fileName) || usedNames.has(name)) {
            name = `Component${++index}`;
            fileName = Path.posix.join(directory, `${name}.vue`);
        }
    }
    return fileName;
}
function indent(code) {
    return '  ' + code.split('\n').join('\n  ');
}

const noop = {
    toggleLineComment(fileName, textRange) {
        return [];
    },
    toggleMultilineComment(fileName, textRange) {
        return [];
    },
    commentSelection(fileName, textRange) {
        return [];
    },
    uncommentSelection(fileName, textRange) {
        return [];
    },
    getRenameInfo(fileName, position, options) {
        return {
            canRename: false,
            localizedErrorMessage: 'Not implemented.',
        };
    },
    findRenameLocations(fileName, position, findInStrings, findInComments) {
        return undefined;
    },
    getApplicableRefactors(fileName, positionOrRange, preferences) {
        return [];
    },
    getBraceMatchingAtPosition(fileName, position) {
        return [];
    },
    getBreakpointStatementAtPosition(fileName, position) {
        return undefined;
    },
    getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
        return [];
    },
    getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
        return {
            changes: [],
        };
    },
    getCompilerOptionsDiagnostics() {
        return [];
    },
    getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
        return undefined;
    },
    getCompletionEntrySymbol(fileName, position, name, source) {
        return undefined;
    },
    getCompletionsAtPosition(fileName, position, options) {
        return undefined;
    },
    getDefinitionAndBoundSpan(fileName, position) {
        return undefined;
    },
    getDefinitionAtPosition(fileName, position) {
        return undefined;
    },
    getDocCommentTemplateAtPosition(fileName, position) {
        return undefined;
    },
    getDocumentHighlights(fileName, position, filesToSearch) {
        return undefined;
    },
    getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
        return [];
    },
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        return undefined;
    },
    getEmitOutput(fileName) {
        return { outputFiles: [], emitSkipped: true };
    },
    getEncodedSemanticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getEncodedSyntacticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getFormattingEditsAfterKeystroke(fileName, position, key, options) {
        return [];
    },
    getFormattingEditsForDocument(fileName, options) {
        return [];
    },
    getFormattingEditsForRange(fileName, start, end, options) {
        return [];
    },
    getImplementationAtPosition(fileName, position) {
        return undefined;
    },
    getIndentationAtPosition(fileName, position, options) {
        return 0;
    },
    getJsxClosingTagAtPosition(fileName, position) {
        return undefined;
    },
    getNameOrDottedNameSpan(fileName, startPos, endPos) {
        return undefined;
    },
    getNavigateToItems(searchValue) {
        return [];
    },
    getNavigationBarItems(fileName) {
        return [];
    },
    getNavigationTree(fileName) {
        return {
            text: '',
            kind: 'module',
            kindModifiers: 'component',
            spans: [],
            nameSpan: undefined,
        };
    },
    getOccurrencesAtPosition(fileName, position) {
        return undefined;
    },
    getOutliningSpans(fileName) {
        return [];
    },
    getQuickInfoAtPosition(fileName, position) {
        return undefined;
    },
    getReferencesAtPosition(fileName, position) {
        return undefined;
    },
    getSemanticClassifications(fileName, span) {
        return [];
    },
    getSemanticDiagnostics(fileName) {
        return [];
    },
    getSignatureHelpItems(fileName, position, options) {
        return undefined;
    },
    getSmartSelectionRange(fileName, position) {
        return {
            textSpan: { start: position, length: 0 },
        };
    },
    getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
        return undefined;
    },
    getSuggestionDiagnostics(fileName) {
        return [];
    },
    getSyntacticClassifications(fileName, span) {
        return [];
    },
    getSyntacticDiagnostics(fileName) {
        return [];
    },
    getTodoComments(fileName, descriptors) {
        return [];
    },
    getTypeDefinitionAtPosition(fileName, position) {
        return [];
    },
    findReferences(fileName, position) {
        return undefined;
    },
    prepareCallHierarchy(fileName, position) {
        return undefined;
    },
    provideCallHierarchyIncomingCalls(fileName, position) {
        return [];
    },
    provideCallHierarchyOutgoingCalls(fileName, position) {
        return [];
    },
    isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
        return false;
    },
    organizeImports(scope, formatOptions, preferences) {
        return [];
    },
    getProgram() {
        throw new Error('Not Implemented');
    },
    cleanupSemanticCache() {
        throw new Error('Not Implemented');
    },
    dispose() {
        throw new Error('Not Implemented');
    },
    applyCodeActionCommand() {
        return Promise.reject('Not Implemented');
    },
};

function createVirtualLanguageServer(config) {
    function removePrefix(fileName) {
        return fileName.startsWith('^vue:') ? fileName.substr(5) : fileName;
    }
    function choose(fileName) {
        var _a, _b;
        return ((_b = (_a = config.context.projectService
            .getDefaultProjectForFile(config.context.typescript.server.toNormalizedPath(fileName), false)) === null || _a === void 0 ? void 0 : _a.getLanguageService()) !== null && _b !== void 0 ? _b : config.service);
    }
    return wrapInTrace('VirtualLanguageServer', {
        ...noop,
        getQuickInfoAtPosition(rawFileName, position) {
            const fileName = removePrefix(rawFileName);
            return choose(fileName).getQuickInfoAtPosition(fileName, position);
        },
    });
}

const GotoExternalComponentProp = {
    version: '*',
    getDefinitionAndBoundSpan({ helpers: h, context }, fileName, position) {
        var _a, _b, _c;
        const document = h.getRenderDoc(fileName);
        if ((document === null || document === void 0 ? void 0 : document.ast) == null)
            return;
        const targetNode = h.findTemplateNodeAt(document.ast, position);
        if (isSimpleExpressionNode(targetNode.node)) {
            const last = targetNode.ancestors.pop();
            if (last != null) {
                targetNode.node = last.node;
            }
        }
        let name;
        let node;
        if (isDirectiveNode(targetNode === null || targetNode === void 0 ? void 0 : targetNode.node) &&
            targetNode.node.name === 'bind' &&
            isSimpleExpressionNode(targetNode.node.arg)) {
            name = targetNode.node.arg.content;
            node = targetNode.node.arg;
        }
        else if (isAttributeNode(targetNode === null || targetNode === void 0 ? void 0 : targetNode.node)) {
            name = targetNode.node.name;
            node = targetNode.node;
        }
        else {
            return;
        }
        const parent = (_a = targetNode.ancestors[targetNode.ancestors.length - 1]) === null || _a === void 0 ? void 0 : _a.node;
        if (!isComponentNode(parent))
            return;
        const componentInfo = h.getComponentInfo(document.container);
        const component = componentInfo.components.find((component) => component.aliases.includes(parent.tag));
        if (component == null)
            return;
        const scriptFileName = ((_b = document.container.getDocument('script')) !== null && _b !== void 0 ? _b : document.container.getDocument('scriptSetup')).fsPath;
        const resolvedModule = h.getResolvedModule(scriptFileName, component.source.moduleName);
        if (resolvedModule == null)
            return;
        if (!isVirtualFile(resolvedModule.resolvedFileName))
            return;
        const externalComponentFilename = getContainingFile(resolvedModule.resolvedFileName);
        const componentDoc = h.getVueDocument(externalComponentFilename);
        if (componentDoc == null)
            return;
        const externalComponentInfo = h.getComponentInfo(componentDoc);
        const prop = externalComponentInfo.props.find((prop) => prop.name === name);
        if (prop == null)
            return;
        return {
            textSpan: {
                start: node.loc.start.offset,
                length: name.length,
            },
            definitions: [
                {
                    name: prop.name,
                    kind: context.typescript.ScriptElementKind.jsxAttribute,
                    fileName: componentDoc.fsPath,
                    textSpan: {
                        start: prop.loc.start.offset,
                        length: prop.loc.source.length,
                    },
                    containerKind: context.typescript.ScriptElementKind.unknown,
                    containerName: '',
                    contextSpan: ((_c = externalComponentInfo.options) === null || _c === void 0 ? void 0 : _c.properties.props) != null
                        ? {
                            start: externalComponentInfo.options.properties.props.loc.start
                                .offset,
                            length: externalComponentInfo.options.properties.props.loc.source
                                .length,
                        }
                        : undefined,
                },
            ],
        };
    },
};

const GotoScriptDefinitions = {
    version: '*',
    getDefinitionAndBoundSpan({ helpers: h, context, service }, fileName, position) {
        var _a;
        const document = h.getRenderDoc(fileName);
        if (document == null)
            return;
        const location = document.getGeneratedOffsetAt(position);
        if (location == null)
            return;
        const result = service.getDefinitionAndBoundSpan(fileName, location.offset);
        if (result == null)
            return;
        let isTextSpanSet = false;
        const definitions = [];
        (_a = result.definitions) === null || _a === void 0 ? void 0 : _a.forEach((definition) => {
            var _a;
            if (h.isRenderFunctionFileName(definition.fileName)) {
                if (definition.fileName !== document.fsPath) {
                    context.log(`Unexpected bound span in ${document.fsPath} resolved to ${definition.fileName}`);
                    return;
                }
                if (!document.isInGeneratedRange(definition.textSpan.start)) {
                    const newResult = service.getDefinitionAndBoundSpan(definition.fileName, definition.textSpan.start);
                    if (newResult != null &&
                        Array.isArray(newResult.definitions) &&
                        newResult.definitions.length > 0) {
                        // TODO: Resolved contextSpan if `newResult` is resolved as return from setup()
                        definitions.push(...newResult.definitions);
                    }
                    else if (definition.kind === 'var') {
                        const componentInfo = h.getComponentInfo(document.container);
                        const prop = componentInfo.props.find((prop) => prop.name === definition.name);
                        if (prop != null) {
                            definitions.push({
                                ...definition,
                                fileName: document.container.fsPath,
                                textSpan: {
                                    start: prop.loc.start.offset,
                                    length: prop.loc.source.length,
                                },
                                containerKind: context.typescript.ScriptElementKind.unknown,
                                containerName: '',
                                contextSpan: ((_a = componentInfo.options) === null || _a === void 0 ? void 0 : _a.properties.props) != null
                                    ? {
                                        start: componentInfo.options.properties.props.loc.start
                                            .offset,
                                        length: componentInfo.options.properties.props.loc.source
                                            .length,
                                    }
                                    : undefined,
                            });
                        }
                    }
                }
                else {
                    const textSpan = h.getTextSpan(document, definition.textSpan);
                    if (textSpan !== definition.textSpan) {
                        isTextSpanSet = true;
                        definition.textSpan = textSpan;
                        if (definition.contextSpan != null) {
                            definition.contextSpan = h.getTextSpan(document, definition.contextSpan);
                        }
                        definition.fileName = document.container.fsPath;
                        definitions.push(definition);
                    }
                }
            }
            else {
                definitions.push(definition);
            }
        });
        result.definitions = definitions;
        if (!isTextSpanSet) {
            const textSpan = getTextSpan(document, position);
            if (textSpan != null)
                result.textSpan = textSpan;
        }
        return result;
    },
};
function getTextSpan(document, position) {
    if (document.ast == null)
        return null;
    const targetNode = findTemplateNodeAt(document.ast, position);
    if ((targetNode === null || targetNode === void 0 ? void 0 : targetNode.node) == null)
        return null;
    if (isAttributeNode(targetNode.node)) {
        return {
            start: targetNode.node.loc.start.offset,
            length: targetNode.node.name.length,
        };
    }
    else if (isElementNode(targetNode.node)) {
        return {
            start: targetNode.node.loc.start.offset + 1,
            length: targetNode.node.tag.length,
        };
    }
    else {
        return {
            start: targetNode.node.loc.start.offset,
            length: targetNode.node.loc.source.length,
        };
    }
}

const GOTO_PROVIDERS = [GotoScriptDefinitions, GotoExternalComponentProp];

const splitRE = /\r?\n/g;
const emptyRE = /^\s*$/;
const needFixRE = /^(\r?\n)*[\t\s]/;
function deIndent(str) {
    if (!needFixRE.test(str)) {
        return str;
    }
    const lines = str.split(splitRE);
    let min = Number.POSITIVE_INFINITY;
    let type;
    let cur;
    let c;
    for (let i = 0; i < lines.length; i++) {
        const line = lines[i];
        if (!emptyRE.test(line)) {
            if (!type) {
                c = line.charAt(0);
                if (c === ' ' || c === '\t') {
                    type = c;
                    cur = count(line, type);
                    if (cur < min) {
                        min = cur;
                    }
                }
                else {
                    return str;
                }
            }
            else {
                cur = count(line, type);
                if (cur < min) {
                    min = cur;
                }
            }
        }
    }
    return lines
        .map(function (line) {
        return line.slice(min);
    })
        .join('\n');
}
function count(line, type) {
    var i = 0;
    while (line.charAt(i) === type) {
        i++;
    }
    return i;
}

/// <reference types="@vuedx/compiler-tsx" />
const RefactorExtractComponent = {
    version: '*',
    findRefactors(config, fileName, position, preferences) {
        const document = config.helpers.getRenderDoc(fileName);
        if (document != null) {
            const nodes = findNodes(config, document, position);
            if (nodes.length !== 0) {
                let isExtractSupported = true;
                const modelExpressions = new Set();
                nodes.forEach((node) => traverseFast(node, (node) => {
                    if (isDirectiveNode(node) && node.name === 'model') {
                        if (isSimpleExpressionNode(node.exp)) {
                            modelExpressions.add(node.exp.content.trim());
                        }
                    }
                }));
                if (modelExpressions.size > 1) {
                    const version = parseInt(config.context.getVueVersion(document.container.fsPath));
                    if (version < 3)
                        isExtractSupported = false;
                }
                if (isExtractSupported) {
                    const extractAsComponent = {
                        name: 'component',
                        description: 'Extract to component',
                        inlineable: true,
                        actions: [
                            {
                                name: ':current',
                                description: 'Extract to component in current directory',
                            },
                        ],
                    };
                    const project = config.context.projectService.getDefaultProjectForFile(config.context.typescript.server.toNormalizedPath(document.fsPath), false);
                    if (project != null) {
                        config.context.config.directories
                            .filter((dir) => dir.kind === 'component')
                            .forEach((config) => {
                            extractAsComponent.actions.push({
                                name: config.name,
                                description: `Extract to component in ${config.name} directory`,
                            });
                        });
                    }
                    if (extractAsComponent.actions.length > 4) {
                        extractAsComponent.inlineable = false;
                    }
                    return [extractAsComponent];
                }
            }
        }
        return [];
    },
    applyRefactor(config, fileName, options, position, refactorName, actionName, preferences) {
        const document = config.helpers.getRenderDoc(fileName);
        if (document != null && refactorName === 'component') {
            const path = Path.posix.dirname(fileName);
            const info = config.helpers.getComponentInfo(document.container);
            const directoryName = path;
            const componentFileName = getFilenameForNewComponent(config.context, directoryName, new Set(info.components.map((component) => component.name)));
            const nodes = findNodes(config, document, position);
            const identifiers = Array.from(new Set(nodes.flatMap((node) => node.scope.identifiers.filter((id) => node.scope.getBinding(id) !== node))));
            const template = document.container
                .getDocument('template')
                .getText()
                .substring(nodes[0].loc.start.offset, nodes[nodes.length - 1].loc.end.offset);
            // TODO: detect types of props.
            const script = [`import { defineComponent } from 'vue'`];
            const components = new Set();
            const models = new Set();
            nodes.forEach((node) => traverseFast(node, (node) => {
                if (isComponentNode(node))
                    components.add(node.tag);
                else if (isDirectiveNode(node) && node.name === 'model') {
                    if (isSimpleExpressionNode(node.exp)) {
                        node.exp.scope.globals.forEach((identifier) => {
                            models.add(identifier);
                        });
                    }
                }
            }));
            const options = {
                components: `components: {${info.components.reduce((text, { name, source }) => {
                    if (components.has(name)) {
                        script.push(`import ${source.exportName != null
                            ? `{ ${source.localName === source.exportName
                                ? ``
                                : `${source.exportName} as `}${source.localName} }`
                            : `${source.localName}`} from '${source.moduleName}'`);
                        return (text +
                            `${name === source.localName
                                ? ``
                                : `${/^[a-z0-9]$/i.test(name) ? `'${name}'` : name}: `}${source.localName}, `);
                    }
                    return text;
                }, '')}},`,
                props: `props: [${identifiers
                    .map((identifier) => `'${identifier}'`)
                    .join(', ')}],`,
            };
            script.push(`export default defineComponent({\n  ${options.props}\n  ${options.components}\n})`);
            const code = `<script>\n${script.join('\n')}\n</script>\n\n<template>\n${formatTemplate(template)}\n</template>\n`;
            let { name, changes, renameLocation, isScript, } = getImportEditForComponent(document.container, info, componentFileName);
            const templateReplacement = {
                newText: `<${name} ${identifiers
                    .map((identifer) => `${models.has(identifer) ? 'v-model' : ''}:${identifer}="${identifer}"`)
                    .join(' ')}/>`,
                span: {
                    start: nodes[0].loc.start.offset,
                    length: template.length,
                },
            };
            changes.push(templateReplacement);
            // Template is before script, need to adjust rename trigger location.
            if (isScript && templateReplacement.span.start < renameLocation) {
                renameLocation =
                    renameLocation - template.length + templateReplacement.newText.length;
            }
            config.context.createVueDocument(componentFileName, code);
            return {
                renameFilename: document.container.fsPath,
                renameLocation,
                edits: [
                    {
                        isNewFile: true,
                        fileName: componentFileName,
                        textChanges: [
                            {
                                span: {
                                    start: 0,
                                    length: 0,
                                },
                                newText: code,
                            },
                        ],
                    },
                    {
                        fileName: document.container.fsPath,
                        textChanges: changes,
                    },
                ],
            };
        }
        return undefined;
    },
};
function formatTemplate(template) {
    const index = template.indexOf('\n');
    if (index >= 0) {
        const firstLine = template.substr(0, index + 1);
        const otherLines = template.substr(index + 1);
        return indent(firstLine + deIndent(otherLines));
    }
    return indent(template);
}
function findNodes(config, document, position) {
    return document.ast != null
        ? typeof position === 'number'
            ? config.helpers.findTemplateNodesIn(document.ast, position, position)
            : config.helpers.findTemplateNodesIn(document.ast, position.pos, position.end)
        : [];
}
function getImportEditForComponent(document, info, fileName) {
    var _a;
    const changes = [];
    const name = getComponentName(fileName);
    const { script, scriptSetup } = document.descriptor;
    const relativeFileName = `./${Path.posix.relative(Path.posix.dirname(document.fsPath), fileName)}`;
    let renameLocation = 0;
    let isScript = false;
    if (scriptSetup != null) {
        const newText = `\nexport { default as ${name} } from '${relativeFileName}';`;
        renameLocation =
            scriptSetup.loc.start.offset +
                newText.indexOf(relativeFileName) +
                relativeFileName.length -
                5;
        changes.push({
            newText,
            span: { start: scriptSetup.loc.start.offset, length: 0 },
        });
    }
    else if (script != null) {
        isScript = true;
        const newText = `\nimport ${name} from '${relativeFileName}';`;
        renameLocation =
            script.loc.start.offset +
                newText.indexOf(relativeFileName) +
                relativeFileName.length -
                5;
        changes.push({
            newText,
            span: { start: script.loc.start.offset, length: 0 },
        });
        if (((_a = info.options) === null || _a === void 0 ? void 0 : _a.properties.components) != null) {
            const components = info.options.properties.components;
            const start = components.loc.start.offset + 1;
            const padding = getPaddingLength(components.loc.source, 1);
            changes.push({
                newText: components.loc.source.substr(1, padding) + `${name},`,
                span: { start, length: 0 },
            });
        }
        else if (info.options != null) {
            const start = info.options.loc.start.offset + 1;
            const padding = getPaddingLength(info.options.loc.source, 1);
            changes.push({
                newText: info.options.loc.source.substr(1, padding) +
                    `components: { ${name} },`,
                span: { start, length: 0 },
            });
        }
        else if (info.setup != null) {
            changes.push({
                newText: `{\n  components: { ${name} },\n setup: `,
                span: { start: info.setup.loc.start.offset, length: 0 },
            }, {
                newText: `\n}`,
                span: { start: info.setup.loc.end.offset, length: 0 },
            });
        }
        else {
            changes.push({
                newText: `\nexport default { components: { ${name} } };\n`,
                span: { start: script.loc.end.offset, length: 0 },
            });
        }
    }
    else {
        isScript = false;
        const newText = `<script>\nimport ${name} from '${relativeFileName}';\nexport default { components: { ${name} } }\n</script>\n\n`;
        renameLocation =
            newText.indexOf(relativeFileName) + relativeFileName.length - 5;
        changes.push({
            newText,
            span: { start: 0, length: 0 },
        });
        // if <script></script>  is there, it should be removed.
        const m = /^<script[^>]*><\/script>/m.exec(document.getText());
        if (m != null) {
            changes.push({
                newText: '',
                span: { start: m.index, length: m[0].length },
            });
        }
    }
    return { name, changes, renameLocation, isScript };
}

const REFACTOR_PROVIDERS = [RefactorExtractComponent];

const RenameExpression = {
    version: '*',
    canRename(config, fileName, position, preferences) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node) && document) {
            const mappedPosition = document.getGeneratedOffsetAt(position);
            if (mappedPosition === null || mappedPosition === void 0 ? void 0 : mappedPosition.offset) {
                //
                const result = config.service.getRenameInfo(fileName, mappedPosition.offset, preferences);
                if (result.canRename) {
                    if (result.displayName === '$event') {
                        return {
                            canRename: false,
                            localizedErrorMessage: '$event is builtin variable, it cannot be renamed.',
                        };
                    }
                    result.triggerSpan = config.helpers.getTextSpan(document, result.triggerSpan);
                }
                return result;
            }
            return {
                canRename: false,
                localizedErrorMessage: 'Cannot find mapped position in render function.',
            };
        }
    },
    applyRename(config, fileName, position, findInStrings, findInComments) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node) && document) {
            const mappedPosition = document.getGeneratedOffsetAt(position);
            if (mappedPosition === null || mappedPosition === void 0 ? void 0 : mappedPosition.offset) {
                return config.service.findRenameLocations(fileName, mappedPosition.offset, findInStrings, findInComments);
            }
        }
    },
    applyFileRename() {
        return undefined;
    },
};

const RenameComponentTag = {
    version: '*',
    canRename(config, fileName, position, options) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (isComponentNode(node) &&
            isPositionInTagName(position, node) &&
            document != null) {
            const info = config.helpers.getComponentInfo(document.container);
            const name = node.tag;
            const component = info.components.find((component) => component.name === name);
            if (component != null &&
                component.kind === 'script' &&
                component.source.exportName != null &&
                component.source.moduleName.endsWith('.vue') &&
                getComponentName(component.source.moduleName) === name) {
                // Ask to rename file instead.
                const newPosition = component.source.loc.start.offset +
                    component.source.loc.source.lastIndexOf(name);
                const result = config.service.getRenameInfo(document.container.getDocumentFileName('script'), newPosition, {
                    allowRenameOfImportPath: true,
                });
                if (result.canRename) {
                    if (node.isSelfClosing ||
                        position < node.loc.start.offset + node.tag.length + 1) {
                        result.triggerSpan = {
                            start: node.loc.start.offset + 1,
                            length: node.tag.length + 4,
                        };
                    }
                    else {
                        result.triggerSpan = {
                            start: node.loc.start.offset + node.loc.source.lastIndexOf('</') + 2,
                            length: node.tag.length + 4,
                        };
                    }
                }
                return result;
            }
            if (component != null) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Global component (with unknown source file) cannot be renamed.',
                };
            }
            return {
                canRename: true,
                displayName: node.tag,
                fullDisplayName: node.tag,
                kind: config.context.typescript.ScriptElementKind.unknown,
                kindModifiers: 'componentTagName',
                fileToRename: undefined,
                triggerSpan: {
                    start: node.loc.start.offset + 1,
                    length: node.tag.length,
                },
            };
        }
    },
    applyRename(config, fileName, position, findInStrings, findInComments) {
        var _a;
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (isComponentNode(node) &&
            isPositionInTagName(position, node) &&
            document != null) {
            const info = config.helpers.getComponentInfo(document.container);
            const name = node.tag;
            const component = info.components.find((component) => component.name === name);
            if ((component === null || component === void 0 ? void 0 : component.kind) === 'script' &&
                component.source.moduleName.endsWith('.vue') &&
                getComponentName(component.source.moduleName) === name) {
                const newPosition = component.source.loc.start.offset +
                    component.source.loc.source.lastIndexOf(name);
                // Ask to rename file instead.
                return (_a = config.service
                    .findRenameLocations(document.container.getDocumentFileName('script'), newPosition, findInStrings, findInComments)) === null || _a === void 0 ? void 0 : _a.slice();
            }
            const renameLocations = [];
            if (component != null) {
                if (component.kind === 'script') {
                    // if component is registered with an alias, then update to reflect in template.
                    if (component.source.localName !== component.name) {
                        renameLocations.push({
                            fileName: document.container.fsPath,
                            textSpan: {
                                start: component.loc.start.offset,
                                length: component.loc.source.length,
                            },
                            ...computeIdentifierReplacement(component.loc.source, component.name),
                        });
                    }
                    // update the import statement.
                    else {
                        const { prefixText } = computeIdentifierReplacement(component.loc.source, component.source.localName);
                        const locations = config.service.findRenameLocations(document.container.getDocumentFileName('script'), component.loc.start.offset + prefixText.length, findInStrings, findInComments);
                        const start = component.source.loc.start.offset;
                        const end = component.source.loc.end.offset;
                        if (locations != null) {
                            const vueFileName = document.container.fsPath;
                            locations.forEach((location) => {
                                if (isVirtualFile(location.fileName) &&
                                    getContainingFile(location.fileName) === vueFileName &&
                                    !location.fileName.endsWith('_render.tsx') &&
                                    (location.textSpan.start < start ||
                                        end < location.textSpan.start)) {
                                    // rename locations except in render file and component registration location.
                                    renameLocations.push(location);
                                }
                            });
                        }
                        const importReplacement = computeIdentifierReplacement(component.source.loc.source, component.source.localName);
                        // add alias to named import.
                        if (component.source.exportName != null &&
                            !importReplacement.prefixText.trim().endsWith(' as')) {
                            importReplacement.prefixText =
                                importReplacement.prefixText +
                                    component.source.exportName +
                                    ' as ';
                        }
                        renameLocations.unshift({
                            fileName: document.container.fsPath,
                            textSpan: {
                                start: component.source.loc.start.offset,
                                length: component.source.loc.source.length,
                            },
                            ...importReplacement,
                        });
                    }
                }
                if (document.ast != null) {
                    traverseFast(document.ast, (node) => {
                        if (isComponentNode(node)) {
                            if (component.aliases.includes(node.tag)) {
                                renameLocations.push({
                                    fileName: document.container.fsPath,
                                    textSpan: {
                                        start: node.loc.start.offset + 1,
                                        length: node.tag.length,
                                    },
                                });
                                if (!node.isSelfClosing) {
                                    renameLocations.push({
                                        fileName: document.container.fsPath,
                                        textSpan: {
                                            start: node.loc.start.offset +
                                                node.loc.source.lastIndexOf('</' + node.tag) +
                                                2,
                                            length: node.tag.length,
                                        },
                                    });
                                }
                            }
                        }
                    });
                }
            }
            return renameLocations;
        }
        return undefined;
    },
    applyFileRename(config, containerFile, oldFileName, newFileName, options, preferences) {
        const fileTextChanges = [];
        if (isVueFile(oldFileName)) {
            const document = config.helpers.getRenderDoc(containerFile);
            if (document != null) {
                const name = getComponentName(oldFileName);
                const newName = getComponentName(newFileName);
                const info = config.helpers.getComponentInfo(document.container);
                const component = info.components.find((component) => component.name === name);
                if (component != null) {
                    const script = document.container.getDocument('script');
                    if (script != null) {
                        const position = Math.max(component.loc.end.offset - 2, component.loc.start.offset);
                        const renameInfo = config.service.getRenameInfo(script.fsPath, position);
                        if (renameInfo.canRename) {
                            const edits = config.service.findRenameLocations(script.fsPath, position, false, false);
                            if (edits != null) {
                                const fileName = document.container.fsPath;
                                fileTextChanges.push({
                                    fileName: fileName,
                                    textChanges: edits
                                        .filter((edit) => isVirtualFile(edit.fileName) &&
                                        getContainingFile(edit.fileName) === fileName)
                                        .map((edit) => ({ span: edit.textSpan, newText: newName })),
                                });
                                const textChanges = [];
                                // update tags in template
                                if (document.ast != null) {
                                    traverseFast(document.ast, (node) => {
                                        if (isComponentNode(node)) {
                                            if (component.aliases.includes(node.tag)) {
                                                textChanges.push({
                                                    newText: newName,
                                                    span: {
                                                        start: node.loc.start.offset + 1,
                                                        length: node.tag.length,
                                                    },
                                                });
                                                if (!node.isSelfClosing) {
                                                    textChanges.push({
                                                        newText: newName,
                                                        span: {
                                                            start: node.loc.start.offset +
                                                                node.loc.source.lastIndexOf('</' + node.tag) +
                                                                2,
                                                            length: node.tag.length,
                                                        },
                                                    });
                                                }
                                            }
                                        }
                                    });
                                }
                                fileTextChanges.push({ fileName, textChanges: textChanges });
                            }
                        }
                    }
                }
            }
        }
        if (fileTextChanges.length > 0)
            return fileTextChanges;
    },
};
function isPositionInTagName(position, node) {
    return (
    // In start tag.
    position <= node.loc.start.offset + node.tag.length ||
        // In end tag.
        (!node.isSelfClosing &&
            node.loc.start.offset + node.loc.source.lastIndexOf('</') <= position));
}

const RenameElementTag = {
    version: '*',
    canRename(config, fileName, position) {
        const { node } = config.helpers.findNodeAtPosition(fileName, position);
        if (isPlainElementNode(node) && isPositionInTagName$1(position, node)) {
            return {
                canRename: true,
                displayName: node.tag,
                fullDisplayName: node.tag,
                kind: config.context.typescript.ScriptElementKind.unknown,
                kindModifiers: 'elementTagName',
                triggerSpan: {
                    start: node.loc.start.offset + 1,
                    length: node.tag.length,
                },
            };
        }
    },
    applyRename(config, fileName, position) {
        const { node, document } = config.helpers.findNodeAtPosition(fileName, position);
        if (isPlainElementNode(node) && isPositionInTagName$1(position, node) && document) {
            const locations = [
                {
                    fileName: document.container.fsPath,
                    textSpan: { start: node.loc.start.offset + 1, length: node.tag.length },
                },
            ];
            if (!node.isSelfClosing) {
                locations.push({
                    fileName: document.container.fsPath,
                    textSpan: {
                        start: node.loc.start.offset + node.loc.source.lastIndexOf('</' + node.tag) + 2,
                        length: node.tag.length,
                    },
                });
            }
            return locations;
        }
    },
    applyFileRename() {
        return undefined;
    },
};
function isPositionInTagName$1(position, node) {
    return (
    // In start tag.
    position <= node.loc.start.offset + node.tag.length ||
        // In end tag.
        (!node.isSelfClosing && node.loc.start.offset + node.loc.source.lastIndexOf('</') <= position));
}

const RENAME_PROVIDERS = [RenameExpression, RenameComponentTag, RenameElementTag];

function createTemplateLanguageServer(config) {
    const { helpers: h, context, service } = config;
    const choose = (fileName) => {
        var _a, _b, _c;
        try {
            (_a = service.getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(fileName); // This should throw if {fileName} is not part of program.
            return service;
        }
        catch (_d) {
            return ((_c = (_b = context.projectService
                .getDefaultProjectForFile(context.typescript.server.toNormalizedPath(fileName), false)) === null || _b === void 0 ? void 0 : _b.getLanguageService()) !== null && _c !== void 0 ? _c : service);
        }
    };
    return wrapInTrace('TemplateLanguageServer', {
        ...noop,
        getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
            var _a, _b, _c;
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return;
            const loc = document.tryGetGeneratedOffset(position);
            const result = (_b = (_a = (loc != null
                ? choose(document.fsPath).getCompletionEntryDetails(document.fsPath, loc, entryName, formatOptions, source, preferences)
                : undefined)) !== null && _a !== void 0 ? _a : choose(document.fsPath).getCompletionEntryDetails(document.fsPath, document.tagCompletionsTriggerOffset, entryName, formatOptions, source, preferences)) !== null && _b !== void 0 ? _b : choose(document.fsPath).getCompletionEntryDetails(document.fsPath, document.contextCompletionsTriggerOffset, entryName, formatOptions, source, preferences);
            (_c = result === null || result === void 0 ? void 0 : result.codeActions) === null || _c === void 0 ? void 0 : _c.forEach((codeAction) => {
                if (source != null &&
                    isVirtualFile(source) &&
                    entryName === 'component') {
                    const { script, scriptSetup } = document.container.descriptor;
                    codeAction.changes.forEach((change) => {
                        const additionalTextChanges = [];
                        if (change.fileName === document.fsPath) {
                            change.fileName = document.container.fsPath;
                            change.textChanges.forEach((textChange) => {
                                if (textChange.newText.startsWith('import component from')) {
                                    if (scriptSetup != null) {
                                        textChange.newText = textChange.newText.replace('import component ', 'export { default as component } ');
                                        textChange.span.start =
                                            scriptSetup.loc.start.offset +
                                                scriptSetup.loc.source.indexOf('\n') +
                                                1; // Insert after first new line
                                        textChange.span.length = 0;
                                    }
                                }
                            });
                        }
                        change.textChanges = [
                            ...change.textChanges,
                            ...additionalTextChanges,
                        ];
                    });
                }
            });
            return result;
        },
        // TODO: Extract this to features directory and modularize.
        getCompletionsAtPosition(fileName, position, options) {
            var _a;
            const document = h.getRenderDoc(fileName);
            const result = {
                isGlobalCompletion: false,
                isMemberCompletion: false,
                isNewIdentifierLocation: false,
                entries: [],
            };
            if (document == null)
                return result;
            const nodeAtCursor = h.findNodeAtPosition(document.fsPath, position);
            const loc = document.tryGetGeneratedOffset(position);
            const { isTagCompletion, isInExpression, charAtCursor, } = detectCompletionType(context, document, position, options === null || options === void 0 ? void 0 : options.triggerCharacter, nodeAtCursor.node);
            if (isTagCompletion) {
                Object.assign(result, choose(document.fsPath).getCompletionsAtPosition(document.fsPath, document.tagCompletionsTriggerOffset, options));
                const info = h.getComponentInfo(document.container);
                info.components.forEach((component) => {
                    var _a;
                    result.entries.push({
                        name: component.name,
                        kind: context.typescript.ScriptElementKind.jsxAttribute,
                        sortText: '4',
                        source: (_a = h.getResolvedModule(document.fsPath, component.source.moduleName)) === null || _a === void 0 ? void 0 : _a.resolvedFileName,
                    });
                });
                // TODO: Add global components
                const project = context.projectService.getDefaultProjectForFile(context.typescript.server.toNormalizedPath(document.fsPath), false);
                if (project != null) {
                    const fileNames = Array.from(new Set(project
                        .getFileNames(false)
                        .map((fileName) => fileName.toString())
                        .filter((fileName) => isVirtualFile(fileName) || isVueFile(fileName))
                        .map((fileName) => getContainingFile(fileName))));
                    fileNames.forEach((fileName) => {
                        const name = getComponentName(fileName);
                        result.entries.push({
                            name: name,
                            kind: context.typescript.ScriptElementKind.jsxAttribute,
                            sortText: '4',
                            source: fileName + VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR,
                            hasAction: true,
                        });
                    });
                }
            }
            else if (loc != null) {
                Object.assign(result, choose(document.fsPath).getCompletionsAtPosition(document.fsPath, loc, options));
            }
            else if (isElementNode(nodeAtCursor.node)) {
                const loc = document.tryGetGeneratedOffset(nodeAtCursor.node.loc.start.offset + 1);
                if (loc != null) {
                    Object.assign(result, choose(document.fsPath).getCompletionsAtPosition(document.fsPath, loc + nodeAtCursor.node.tag.length + 1, // TODO: Handle kebab-cased tags.
                    options));
                }
            }
            else if (isDirectiveNode(nodeAtCursor.node)) {
                const parent = nodeAtCursor.ancestors[nodeAtCursor.ancestors.length - 1].node;
                if (isElementNode(parent)) {
                    const loc = document.tryGetGeneratedOffset(parent.loc.start.offset + 1);
                    if (loc != null) {
                        Object.assign(result, choose(document.fsPath).getCompletionsAtPosition(document.fsPath, loc + parent.tag.length + 1, // TODO: Handle kebab-cased tags.
                        options));
                    }
                }
            }
            if (isInExpression && /^[ {+/*-]$/.test(charAtCursor)) {
                const contextCompletion = choose(document.fsPath).getCompletionsAtPosition(document.fsPath, document.contextCompletionsTriggerOffset, {
                    ...options,
                    triggerCharacter: '.',
                    includeAutomaticOptionalChainCompletions: true,
                    includeCompletionsForModuleExports: false,
                    includeCompletionsWithInsertText: true,
                    includePackageJsonAutoImports: 'off',
                    provideRefactorNotApplicableReason: false,
                    disableSuggestions: true,
                });
                if ((contextCompletion === null || contextCompletion === void 0 ? void 0 : contextCompletion.entries) != null) {
                    result.entries = (_a = result.entries) !== null && _a !== void 0 ? _a : [];
                    result.entries.push(...contextCompletion.entries);
                }
            }
            if ((result === null || result === void 0 ? void 0 : result.entries) != null) {
                const disallowedIdentifiers = new Set([
                    'arguments',
                    'globalThis',
                    'default',
                ]);
                const allowedGlobals = new Set([
                    'Infinity',
                    'undefined',
                    'NaN',
                    'isFinite',
                    'isNaN',
                    'parseFloat',
                    'parseInt',
                    'decodeURI',
                    'decodeURIComponent',
                    'encodeURI',
                    'encodeURIComponent',
                    'Math',
                    'Number',
                    'Date',
                    'Array',
                    'Object',
                    'Boolean',
                    'String',
                    'RegExp',
                    'Map',
                    'Set',
                    'JSON',
                    'Intl',
                ]);
                if (isInExpression) {
                    const { components } = h.getComponentInfo(document.container);
                    components.forEach((component) => component.aliases.forEach((alias) => disallowedIdentifiers.add(alias)));
                }
                result.entries = result.entries.filter((entry) => {
                    if (entry.source != null && entry.kind !== 'JSX attribute')
                        return; // Ignore external module import
                    if (disallowedIdentifiers.has(entry.name))
                        return false;
                    if (entry.name.startsWith('_'))
                        return false; // Ignore Vue internals
                    if (entry.kindModifiers != null) {
                        if (entry.kindModifiers.includes('export')) {
                            if (entry.kind !== 'property')
                                return false; // Ignore non-property exports
                        }
                        if (entry.kindModifiers.includes('deprecated'))
                            return false; // Ignore deprecated
                        if (entry.kindModifiers.includes('declare')) {
                            if (entry.kind !== 'property') {
                                if (!allowedGlobals.has(entry.name))
                                    return false;
                            }
                        }
                    }
                    if (entry.source === 'constants')
                        return false; // Ignore typescript constants
                    if (entry.kind === 'keyword')
                        return false; // Only helpful in v-on but we discourage big inline handlers.
                    if (entry.name.startsWith('$')) {
                        entry.sortText = '9';
                    }
                    if (isTagCompletion && entry.kind !== 'JSX attribute') {
                        return false;
                    }
                    return true;
                });
            }
            return result;
        },
        getSignatureHelpItems(fileName, position, options) {
            const document = h.getRenderDoc(fileName);
            const loc = document === null || document === void 0 ? void 0 : document.getGeneratedOffsetAt(position);
            if (loc != null && document != null) {
                return choose(document.fsPath).getSignatureHelpItems(document.fsPath, loc.offset, options);
            }
        },
        getQuickInfoAtPosition(fileName, position) {
            var _a;
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return;
            const loc = document.getGeneratedOffsetAt(position);
            if (loc == null)
                return;
            const result = choose(document.fsPath).getQuickInfoAtPosition(fileName, loc.offset);
            if (result != null) {
                const textSpan = h.getTextSpan(document, result.textSpan);
                if (textSpan != null) {
                    result.textSpan = textSpan;
                }
                else {
                    return;
                }
                (_a = result.displayParts) === null || _a === void 0 ? void 0 : _a.forEach((displayPart) => {
                    if (displayPart.text === 'JSX attribute') {
                        displayPart.text = 'prop'; // TODO: Maybe event or prop or attribute?
                    }
                });
            }
            return result;
        },
        getSemanticDiagnostics(fileName) {
            var _a;
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return [];
            const diagnostics = choose(document.fsPath)
                .getSemanticDiagnostics(document.fsPath)
                .map((diagnostic) => {
                var _a;
                if (diagnostic.file != null &&
                    diagnostic.file.fileName !== document.fsPath) {
                    return diagnostic;
                }
                if (diagnostic.start != null) {
                    if (document.isInGeneratedRange(diagnostic.start)) {
                        const position = document.getOriginalOffsetAt(diagnostic.start);
                        diagnostic.start = (_a = position === null || position === void 0 ? void 0 : position.offset) !== null && _a !== void 0 ? _a : diagnostic.start;
                        return diagnostic;
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return diagnostic;
                }
            })
                .filter(isNotNull);
            // TODO: Cache it.
            const project = context.getVueProjectForFile(document.container.fsPath);
            if (document.ast != null && (project === null || project === void 0 ? void 0 : project.kind) === 'inferred') {
                const info = h.getComponentInfo(document.container);
                const localComponentNames = new Set(info.components.flatMap((component) => component.aliases));
                const globalComponentNames = new Set(project.globalComponents.flatMap((component) => component.aliases));
                const projectComponentNames = new Set(project.components.flatMap((component) => component.aliases));
                const projectComponents = project.components;
                const file = (_a = choose(document.fsPath)
                    .getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(document.fsPath);
                traverseFast(document.ast, (node) => {
                    if (isComponentNode(node)) {
                        if (!localComponentNames.has(node.tag) &&
                            !globalComponentNames.has(node.tag) &&
                            projectComponentNames.has(node.tag)) {
                            const components = projectComponents.filter((component) => component.aliases.includes(node.tag));
                            let messageText = `The component '${node.tag}' is inferred as global component. It may not be available at runtime.`;
                            const relatedInformation = [];
                            if (!components[0].source.moduleName.endsWith('.vue')) {
                                messageText = `The component '${node.tag}' is found in '${components[0].source.moduleName}' and it inferred as global component. It may not be available at runtime.`;
                            }
                            if (components.length > 1) {
                                components.forEach((component) => {
                                    relatedInformation.push({
                                        code: 59002,
                                        category: context.typescript.DiagnosticCategory.Warning,
                                        messageText: `Found in '${component.source.moduleName}'`,
                                        file: undefined,
                                        start: undefined,
                                        length: undefined,
                                    });
                                });
                            }
                            diagnostics.push({
                                category: context.typescript.DiagnosticCategory.Warning,
                                code: 59001,
                                start: node.loc.start.offset + 1,
                                length: node.tag.length,
                                messageText,
                                file,
                                relatedInformation,
                            });
                        }
                    }
                });
            }
            return diagnostics;
        },
        getSuggestionDiagnostics(fileName) {
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return [];
            const diagnostics = choose(document.fsPath)
                .getSuggestionDiagnostics(document.fsPath)
                .map((diagnostic) => {
                var _a;
                if (diagnostic.start != null) {
                    if (document.isInGeneratedRange(diagnostic.start)) {
                        const position = document.getOriginalOffsetAt(diagnostic.start);
                        diagnostic.start = (_a = position === null || position === void 0 ? void 0 : position.offset) !== null && _a !== void 0 ? _a : diagnostic.start;
                        return diagnostic;
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return diagnostic;
                }
            })
                .filter(isNotNull);
            return diagnostics;
        },
        getSyntacticDiagnostics(fileName) {
            var _a;
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return [];
            const diagnostics = choose(document.fsPath)
                .getSyntacticDiagnostics(document.fsPath)
                .map((diagnostic) => {
                var _a;
                if (diagnostic.start != null) {
                    if (document.isInGeneratedRange(diagnostic.start)) {
                        const position = document.getOriginalOffsetAt(diagnostic.start);
                        diagnostic.start = (_a = position === null || position === void 0 ? void 0 : position.offset) !== null && _a !== void 0 ? _a : diagnostic.start;
                        return diagnostic;
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return diagnostic;
                }
            })
                .filter(isNotNull);
            if (document.parserErrors.length > 0) {
                const sourceFile = (_a = choose(document.fsPath)
                    .getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(document.fsPath);
                const block = document.container.getBlock({ type: 'template' });
                if (sourceFile != null && block != null) {
                    const start = block.loc.start.offset;
                    const length = block.loc.end.offset - start;
                    document.parserErrors.forEach((error) => {
                        var _a;
                        diagnostics.push({
                            category: context.typescript.DiagnosticCategory.Error,
                            code: 50000 + error.code,
                            file: sourceFile,
                            source: (_a = error.loc) === null || _a === void 0 ? void 0 : _a.source,
                            start: error.loc != null ? error.loc.start.offset : start,
                            length: error.loc != null
                                ? error.loc.end.offset - error.loc.start.offset
                                : length,
                            messageText: error.message,
                        });
                    });
                }
            }
            return diagnostics;
        },
        getRenameInfo(fileName, position, preferences = {}) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.canRename(config, fileName, position, preferences);
                if (result != null)
                    return result;
            }
            return {
                canRename: false,
                localizedErrorMessage: 'You cannot rename this element.',
            };
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.applyRename(config, fileName, position, findInStrings, findInComments);
                if (result != null)
                    return result;
            }
            return [];
        },
        getEditsForFileRenameIn(fileName, oldFilePath, newFilePath) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.applyFileRename(config, fileName, oldFilePath, newFilePath, {}, {});
                if (result != null)
                    return result;
            }
            return [];
        },
        getApplicableRefactors(fileName, position, preferences = {}) {
            const refactors = [];
            for (const provider of REFACTOR_PROVIDERS) {
                const result = provider.findRefactors(config, fileName, position, preferences);
                if (result != null)
                    refactors.push(...result);
            }
            return refactors;
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences = {}) {
            for (const provider of REFACTOR_PROVIDERS) {
                const result = provider.applyRefactor(config, fileName, formatOptions, positionOrRange, refactorName, actionName, preferences);
                if (result != null)
                    return result;
            }
        },
        getDefinitionAndBoundSpan(fileName, position) {
            for (const provider of GOTO_PROVIDERS) {
                const result = provider.getDefinitionAndBoundSpan(config, fileName, position);
                if (result != null)
                    return result;
            }
        },
    });
}
function detectCompletionType(context, document, position, triggerCharacter, node) {
    const templateSource = document.container.getDocument('template').getText();
    let isTagCompletion = triggerCharacter === '<';
    const index = isTagCompletion
        ? 0
        : templateSource.substr(0, position).lastIndexOf('<');
    if (index > 0) {
        const text = templateSource.substring(index, position);
        isTagCompletion = /^<[A-Za-z0-9-]*$/.test(text);
    }
    const isInExpression = !isTagCompletion &&
        (isSimpleExpressionNode(node) || isInterpolationNode(node));
    return {
        isTagCompletion,
        isInExpression,
        charAtCursor: templateSource.substr(position - 1, 1),
    };
}

function createVueLanguageServer(options) {
    const template = createTemplateLanguageServer(options);
    const { helpers: h, service, context } = options;
    const script = wrapInTrace('ScriptLanguageServer', service);
    function isFeatureEnabled(featureName, checkFor = true) {
        const feature = context.config.features[featureName];
        return Array.isArray(feature)
            ? feature.includes(checkFor)
            : feature === checkFor;
    }
    function choose(document) {
        if (h.isRenderFunctionDocument(document)) {
            return template;
        }
        return script;
    }
    return wrapInTrace('VueLanguageServer', {
        ...noop,
        getSemanticDiagnostics(fileName) {
            if (!isFeatureEnabled('diagnostics', 'semantic')) {
                return [];
            }
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document != null) {
                const selectors = ['script', '_render', 'scriptSetup'];
                selectors.forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual != null) {
                        const results = choose(virtual).getSemanticDiagnostics(virtual.fsPath);
                        diagnostics.push(...results);
                    }
                });
            }
            return diagnostics;
        },
        getSuggestionDiagnostics(fileName) {
            if (!isFeatureEnabled('diagnostics', 'suggestion')) {
                return [];
            }
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document != null) {
                const selectors = ['script', '_render', 'scriptSetup'];
                selectors.forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual != null)
                        diagnostics.push(...choose(virtual).getSuggestionDiagnostics(virtual.fsPath));
                });
            }
            return diagnostics;
        },
        getSyntacticDiagnostics(fileName) {
            if (!isFeatureEnabled('diagnostics', 'syntactic')) {
                return [];
            }
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document != null) {
                const selectors = ['script', '_render', 'scriptSetup'];
                selectors.forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual != null)
                        diagnostics.push(...choose(virtual).getSyntacticDiagnostics(virtual.fsPath));
                });
            }
            return diagnostics;
        },
        organizeImports(scope, formatOptions, preferences) {
            var _a;
            if (!isFeatureEnabled('organizeImports'))
                return [];
            const document = h.getVueDocument(scope.fileName);
            if (document != null) {
                const virtual = (_a = document.getDocument('script')) !== null && _a !== void 0 ? _a : document.getDocument('scriptSetup');
                if (virtual != null) {
                    return script.organizeImports({ ...scope, fileName: virtual.fsPath }, formatOptions, preferences);
                }
            }
            return [];
        },
        getQuickInfoAtPosition(fileName, position) {
            if (!isFeatureEnabled('quickInfo'))
                return;
            // TODO: Provide better quick info for components and props.
            const document = h.getDocumentAt(fileName, position);
            if (document != null)
                return choose(document).getQuickInfoAtPosition(document.fsPath, position);
        },
        getRenameInfo(fileName, position, options) {
            if (!isFeatureEnabled('rename')) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Rename feature disabled.',
                };
            }
            const document = h.getDocumentAt(fileName, position);
            if (document == null) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Cannot find this Vue file.',
                };
            }
            return choose(document).getRenameInfo(document.fsPath, position, options);
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            if (!isFeatureEnabled('rename'))
                return;
            const document = h.getDocumentAt(fileName, position);
            if (document == null)
                return;
            return choose(document).findRenameLocations(document.fsPath, position, findInStrings, findInComments);
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            if (!isFeatureEnabled('rename'))
                return [];
            const document = h.getVueDocument(oldFilePath);
            const fileTextChanges = [];
            const visited = new Set();
            if (document != null) {
                const component = document.getDocument(MODULE_SELECTOR);
                const currentChanges = script.getEditsForFileRename(component.fsPath, isVueFile(newFilePath)
                    ? newFilePath + VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR
                    : newFilePath, formatOptions, preferences);
                fileTextChanges.push(...currentChanges);
                currentChanges.forEach((item) => {
                    var _a;
                    if (isVirtualFile(item.fileName) || isVueFile(item.fileName)) {
                        const render = (_a = h
                            .getVueDocument(item.fileName)) === null || _a === void 0 ? void 0 : _a.getDocument('_render');
                        if (render != null && !visited.has(render.fsPath)) {
                            visited.add(render.fsPath);
                            fileTextChanges.push(...template.getEditsForFileRenameIn(render.fsPath, oldFilePath, newFilePath));
                        }
                    }
                });
            }
            return fileTextChanges;
        },
        getApplicableRefactors(fileName, positionOrRange, preferences) {
            if (!isFeatureEnabled('refactor'))
                return [];
            if (isNumber(positionOrRange)) {
                const document = h.getDocumentAt(fileName, positionOrRange);
                if (document == null)
                    return [];
                return choose(document).getApplicableRefactors(document.fsPath, positionOrRange, preferences);
            }
            else {
                const document1 = h.getDocumentAt(fileName, positionOrRange.pos);
                const document2 = h.getDocumentAt(fileName, positionOrRange.end);
                if (document1 == null || document1 !== document2)
                    return [];
                return choose(document1).getApplicableRefactors(document1.fsPath, positionOrRange, preferences);
            }
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
            if (!isFeatureEnabled('refactor'))
                return;
            const document = h.getDocumentAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos);
            const document2 = h.getDocumentAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.end);
            if (document != null && document === document2) {
                return choose(document).getEditsForRefactor(document.fsPath, formatOptions, positionOrRange, refactorName, actionName, preferences);
            }
        },
        getDocumentHighlights(fileName, position, filesToSearch) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null && !h.isRenderFunctionDocument(document)) {
                return script.getDocumentHighlights(document.fsPath, position, filesToSearch
                    .map((fileName) => {
                    var _a;
                    if (isVueFile(fileName)) {
                        const document = h.getVueDocument(fileName);
                        if (document == null)
                            return;
                        const script = (_a = document.getDocument('script')) !== null && _a !== void 0 ? _a : document.getDocument('scriptSetup');
                        if (script == null)
                            return;
                        return script.fsPath;
                    }
                    return fileName;
                })
                    .filter(Boolean));
            }
        },
        getCompletionsAtPosition(fileName, position, options) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getCompletionsAtPosition(document.fsPath, position, options);
            }
        },
        getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getCompletionEntryDetails(document.fsPath, position, entryName, formatOptions, source, preferences);
            }
        },
        getSignatureHelpItems(fileName, position, options) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getSignatureHelpItems(document.fsPath, position, options);
            }
        },
        getCompletionEntrySymbol(fileName, position, name, source) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getCompletionEntrySymbol(document.fsPath, position, name, source);
            }
        },
        getDefinitionAtPosition(fileName, position) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getDefinitionAtPosition(document.fsPath, position);
            }
        },
        getDefinitionAndBoundSpan(fileName, position) {
            if (!isFeatureEnabled('goto'))
                return;
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getDefinitionAndBoundSpan(document.fsPath, position);
            }
        },
    });
}

const VUE_LANGUAGE_SERVER = Symbol('Vue Language Server');
class RoutingLanguageServer {
    constructor(context) {
        this.context = context;
    }
    decorate(languageService) {
        if (VUE_LANGUAGE_SERVER in languageService) {
            return languageService;
        }
        const proxy = createLanguageServiceRouter({
            context: this.context,
            service: languageService,
            helpers: createServerHelper(this.context, languageService),
        });
        // @ts-expect-error
        proxy[VUE_LANGUAGE_SERVER] = true;
        return proxy;
    }
}
function createLanguageServiceRouter(config) {
    const vue = createVueLanguageServer(config);
    const virtual = createVirtualLanguageServer(config);
    const ts = config.service;
    function choose(fileName) {
        return fileName.startsWith('^vue:')
            ? virtual
            : isVueFile(fileName) || isVirtualFile(fileName)
                ? vue
                : ts;
    }
    function getTextSpan(document, span) {
        if (config.helpers.isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result != null)
                return { start: result.offset, length: result.length };
            return null;
        }
        return span;
    }
    const VIRTUAL_FILE_SUFFIX_RE = new RegExp(`(?<=\\.vue)${VIRTUAL_FILENAME_SEPARATOR}([A-Za-z_][A-Za-z0-9_-]*)(\\.[jt]sx?)?`, 'g');
    function getComponentName$1(fileName) {
        if (fileName == null || !isVueFile(fileName))
            return;
        return getComponentName(fileName);
    }
    const COMPONENT_TYPE_RE = /'ComponentPublicInstance<.*?ComponentOptionsBase<...>>'/g;
    const REPLACE = {
        virtualFile: (messageText) => messageText.replace(VIRTUAL_FILE_SUFFIX_RE, ''),
        componentType: (fileName, messageText) => {
            const component = getComponentName$1(fileName);
            return component != null
                ? messageText.replace(COMPONENT_TYPE_RE, `vue component '${component}'`)
                : messageText;
        },
    };
    function applyReplacements(fileName, messageText) {
        if (typeof messageText === 'string') {
            return REPLACE.componentType(fileName, REPLACE.virtualFile(messageText));
        }
        else {
            messageText.messageText = applyReplacements(fileName, messageText.messageText);
            return messageText;
        }
    }
    function isVirtualSourceFile(file) {
        return file != null && isVirtualFile(file.fileName);
    }
    const proxy = wrapInTrace('RoutingLanguageServer', {
        ...config.service,
        dispose() {
            config.context.disposeUnusedProjects(true);
            config.service.dispose();
        },
        getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
            return config.service.getCombinedCodeFix(scope, fixId, formatOptions, preferences);
        },
        getProgram() {
            return config.service.getProgram();
        },
        toLineColumnOffset(fileName, position) {
            return config.service.toLineColumnOffset != null
                ? config.service.toLineColumnOffset(fileName, position)
                : { line: 0, character: 0 };
        },
        organizeImports(scope, formatOptions, preferences) {
            return choose(scope.fileName)
                .organizeImports(scope, formatOptions, preferences)
                .map((change) => {
                if (isVirtualFile(change.fileName))
                    change.fileName = getContainingFile(change.fileName);
                return change;
            });
        },
        getQuickInfoAtPosition(fileName, position) {
            const info = choose(fileName).getQuickInfoAtPosition(fileName, position);
            if ((info === null || info === void 0 ? void 0 : info.displayParts) != null) {
                info.displayParts = info.displayParts
                    .map((part) => {
                    part.text = applyReplacements(fileName, part.text);
                    return part;
                })
                    .filter(isNotNull);
            }
            if ((info === null || info === void 0 ? void 0 : info.documentation) != null) {
                info.documentation = info.documentation
                    .map((part) => {
                    part.text = applyReplacements(fileName, part.text);
                    return part;
                })
                    .filter(isNotNull);
            }
            return info;
        },
        getSemanticDiagnostics(fileName) {
            config.context.disposeUnusedProjects();
            const diagnostics = choose(fileName).getSemanticDiagnostics(fileName);
            return diagnostics
                .map((diagnostic) => {
                if (isVirtualSourceFile(diagnostic.file)) {
                    diagnostic.file = {
                        ...diagnostic.file,
                        fileName: getContainingFile(diagnostic.file.fileName),
                    };
                }
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation != null) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (info.file != null && isVirtualFile(info.file.fileName)) {
                            info.file = {
                                ...info.file,
                                fileName: getContainingFile(info.file.fileName),
                            };
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getSyntacticDiagnostics(fileName) {
            const diagnostics = choose(fileName).getSyntacticDiagnostics(fileName);
            return diagnostics
                .map((diagnostic) => {
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation != null) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (info.file != null && isVirtualFile(info.file.fileName)) {
                            info.file = {
                                ...info.file,
                                fileName: getContainingFile(info.file.fileName),
                            };
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getCompletionsAtPosition(fileName, position, options) {
            const completions = choose(fileName).getCompletionsAtPosition(fileName, position, options);
            if ((completions === null || completions === void 0 ? void 0 : completions.entries) != null) {
                completions.entries = completions.entries
                    .filter((entry) => {
                    if (entry.source != null && isVirtualFile(entry.source)) {
                        return entry.source.endsWith(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR);
                    }
                    return true;
                })
                    .map((entry) => {
                    if (entry.source != null && isVirtualFile(entry.source)) {
                        entry.source = getContainingFile(entry.source);
                        if (entry.name === 'component') {
                            entry.name = getComponentName(entry.source);
                        }
                    }
                    return entry;
                });
            }
            return completions;
        },
        getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
            var _a, _b;
            const originalEntryName = entryName;
            const originalSource = source;
            if (source != null && isVueFile(source)) {
                const componentName = getComponentName(source);
                source = source + VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR;
                if (entryName === componentName) {
                    entryName = 'component';
                }
            }
            const details = choose(fileName).getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences);
            if (originalSource != null && isVueFile(originalSource)) {
                (_a = details === null || details === void 0 ? void 0 : details.codeActions) === null || _a === void 0 ? void 0 : _a.forEach((codeAction) => {
                    const isComponentImport = entryName === 'component';
                    if (isComponentImport) {
                        codeAction.description = codeAction.description
                            .replace(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR, '')
                            .replace('component', originalEntryName);
                    }
                    codeAction.changes.forEach((change) => {
                        if (isVirtualFile(change.fileName)) {
                            change.fileName = getContainingFile(change.fileName);
                        }
                        change.textChanges.forEach((textChange) => {
                            textChange.newText = textChange.newText.replace(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR, '');
                            if (isComponentImport) {
                                textChange.newText = textChange.newText.replace(/\bcomponent\b/, originalEntryName);
                            }
                        });
                    });
                });
                (_b = details === null || details === void 0 ? void 0 : details.source) === null || _b === void 0 ? void 0 : _b.forEach((item) => {
                    if (isVirtualFile(item.text)) {
                        item.text = getContainingFile(item.text);
                    }
                });
            }
            return details;
        },
        getSignatureHelpItems(fileName, position, options) {
            return choose(fileName).getSignatureHelpItems(fileName, position, options);
        },
        getSuggestionDiagnostics(fileName) {
            const diagnostics = choose(fileName).getSuggestionDiagnostics(fileName);
            return diagnostics
                .map((diagnostic) => {
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation != null) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (info.file != null && isVirtualFile(info.file.fileName)) {
                            info.file = {
                                ...info.file,
                                fileName: getContainingFile(info.file.fileName),
                            };
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
            return choose(fileName).getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences);
        },
        getRenameInfo(fileName, position, options) {
            const result = choose(fileName).getRenameInfo(fileName, position, options);
            if (result.canRename) {
                if (result.fileToRename != null &&
                    isVirtualFile(result.fileToRename)) {
                    result.fileToRename = getContainingFile(result.fileToRename);
                }
                result.displayName = REPLACE.virtualFile(result.displayName);
                result.fullDisplayName = REPLACE.virtualFile(result.fullDisplayName);
                if (result.kind ===
                    config.context.typescript.ScriptElementKind.moduleElement &&
                    result.displayName.endsWith('.vue')) {
                    result.triggerSpan.length -= 4;
                }
            }
            return result;
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            var _a;
            const result = (_a = choose(fileName)
                .findRenameLocations(fileName, position, findInStrings, findInComments)) === null || _a === void 0 ? void 0 : _a.map((item) => {
                if (isVirtualFile(item.fileName)) {
                    const virtual = config.helpers.getDocument(item.fileName);
                    item.fileName = virtual.container.fsPath;
                    const textSpan = getTextSpan(virtual, item.textSpan);
                    if (textSpan == null)
                        return;
                    item.textSpan = textSpan;
                    if (item.contextSpan != null) {
                        const contextSpan = getTextSpan(virtual, item.contextSpan);
                        if (contextSpan == null)
                            return;
                        item.contextSpan = contextSpan;
                    }
                }
                return item;
            }).filter(isNotNull);
            return result;
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            const suffix = '.vue' + VIRTUAL_FILENAME_SEPARATOR + '_module';
            return choose(oldFilePath)
                .getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences)
                .map((edit) => {
                if (isVirtualFile(edit.fileName)) {
                    const result = parseVirtualFileName(edit.fileName);
                    if (result == null)
                        return;
                    if (!['script', 'scriptSetup'].includes(result.selector.type))
                        return;
                    edit.fileName = getContainingFile(edit.fileName);
                }
                edit.textChanges = edit.textChanges
                    .map((change) => {
                    const index = change.newText.indexOf(suffix);
                    if (index >= 0) {
                        change.newText =
                            change.newText.substr(0, index + 4) +
                                change.newText.substr(index + suffix.length);
                    }
                    return change;
                })
                    .filter(isNotNull);
                if (edit.textChanges.length === 0)
                    return;
                return edit;
            })
                .filter(isNotNull);
        },
        getApplicableRefactors(fileName, positionOrRange, preferences) {
            return choose(fileName).getApplicableRefactors(fileName, positionOrRange, preferences);
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
            const result = choose(fileName).getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences);
            const editsByFileName = {};
            if (result != null) {
                result.edits = result.edits.filter((edit) => {
                    var _a;
                    if (isVirtualFile(edit.fileName) || isVueFile(edit.fileName)) {
                        edit.fileName = getContainingFile(edit.fileName);
                        if (!(edit.fileName in editsByFileName)) {
                            editsByFileName[edit.fileName] = {
                                isNewFile: false,
                                fileName: edit.fileName,
                                textChanges: [],
                            };
                        }
                        editsByFileName[edit.fileName].textChanges.push(...edit.textChanges);
                        editsByFileName[edit.fileName].isNewFile = (_a = editsByFileName[edit.fileName].isNewFile) !== null && _a !== void 0 ? _a : edit.isNewFile === true;
                        return false;
                    }
                    return true;
                });
                result.edits.push(...Object.values(editsByFileName));
            }
            config.context.log(JSON.stringify(result, null, 2));
            return result;
        },
        getDefinitionAndBoundSpan(fileName, position) {
            const result = choose(fileName).getDefinitionAndBoundSpan(fileName, position);
            if ((result === null || result === void 0 ? void 0 : result.definitions) != null) {
                result.definitions.forEach((definition) => {
                    var _a, _b;
                    if (isVirtualFile(definition.fileName)) {
                        if (config.helpers.isVueModuleFileName(definition.fileName)) {
                            const document = config.helpers.getVueDocument(definition.fileName);
                            if ((document === null || document === void 0 ? void 0 : document.descriptor) != null) {
                                const script = (_a = document.descriptor.scriptSetup) !== null && _a !== void 0 ? _a : document.descriptor.script;
                                if (script != null) {
                                    // TODO: resolve to export default in <script> or <script setup>
                                    definition.name = getComponentName(document.fsPath);
                                    const info = config.helpers.getComponentInfo(document);
                                    definition.textSpan = definition.contextSpan = {
                                        start: script.loc.start.offset,
                                        length: script.loc.end.offset - script.loc.start.offset,
                                    };
                                    if (((_b = info.options) === null || _b === void 0 ? void 0 : _b.loc) != null) {
                                        definition.textSpan = {
                                            start: info.options.loc.start.offset,
                                            length: info.options.loc.end.offset -
                                                info.options.loc.start.offset +
                                                1,
                                        };
                                    }
                                }
                            }
                        }
                        definition.fileName = getContainingFile(definition.fileName);
                    }
                });
            }
            return result;
        },
        getDefinitionAtPosition(fileName, position) {
            const result = choose(fileName).getDefinitionAtPosition(fileName, position);
            if (result != null) {
                result.forEach((definitionInfo) => {
                    if (isVirtualFile(definitionInfo.fileName)) {
                        definitionInfo.fileName = getContainingFile(definitionInfo.fileName);
                        definitionInfo.name = applyReplacements(definitionInfo.fileName, definitionInfo.name);
                    }
                });
            }
            return result;
        },
        getDocumentHighlights(fileName, position, filesToSearch) {
            const result = choose(fileName).getDocumentHighlights(fileName, position, filesToSearch);
            if (result != null) {
                result.forEach((highlights) => {
                    if (isVirtualFile(fileName)) {
                        highlights.fileName = getContainingFile(highlights.fileName);
                        highlights.highlightSpans.forEach((span) => {
                            if (span.fileName != null && isVirtualFile(span.fileName)) {
                                span.fileName = getContainingFile(span.fileName);
                            }
                        });
                    }
                });
            }
            return result !== null && result !== void 0 ? result : [];
        },
    });
    return proxy;
}

let context;
let server;
function init({ typescript }) {
    context = context !== null && context !== void 0 ? context : new PluginContext(typescript);
    server = server !== null && server !== void 0 ? server : new RoutingLanguageServer(context);
    return {
        create(info) {
            context.load(info);
            return server.decorate(info.languageService);
        },
        getExternalFiles(project) {
            var _a;
            return (_a = context.getExternalFiles(project)) !== null && _a !== void 0 ? _a : [];
        },
        onConfigurationChanged(config) {
            context.setConfig(config);
        },
    };
}

export default init;
//# sourceMappingURL=index.esm.js.map
